package provide tclcron 1.0

namespace eval tclcron {}

# initialize variables
if {![info exists tclcron::jobs]} {
    set tclcron::jobs [list]
}

if {![info exists tclcron::jobId]} {
    set tclcron::jobId 0
}

if {![info exists tclcron::maxWait]} {
    set tclcron::maxWait 2
}

if {![info exists tclcron::after]} {
    set tclcron::after ""
}

# public API

# tclcron::schedule --
#
#	Schedule a command execution at specified interval or at specified times.
#
# Arguments:
#	command	command to be executed.
#	type	type of scheduling (ie 'every').
#	args	type-specific arguments.
#
# Results:
#	Identified of the newly created scheduled procedure.


proc tclcron::schedule {command type args} {
    variable jobs; variable jobId
    
    package require tclcron::tclcron_$type
    
    
    set idx [incr jobId]
    set id [llength $jobs]
    lappend jobs [list $idx {} $type [tclcron::${type}::initialize $idx $args] $command true]

    _reschedule $id
    return $idx
}

# tclcron::unschedule --
#
#	Unschedule a previously scheduled procedure, by it's identifier.
#
# Arguments:
#	identifier	Identifier of the command to remove.
#
# Results:
#	none

proc tclcron::unschedule {identifier} {
    variable jobs
    set temp $jobs
    set rc [list]
    
    foreach job $jobs {
	if {[lindex $job 0] == $identifier} {
	    continue
	}
	lappend rc $job
    }

    set jobs $rc
    return ""
}

# tclcron::info --
#
#	Retrieve a list of all commands, or by a specified pattern.
#
# Arguments:
#	commandPattern	(optional) pattern to match against commands.
#
# Results:
#	A list of all commands, with the following elements:
#	jobID, command, type, typeArguments


proc tclcron::info {{commandPattern *}} {
    variable jobs

    set rc [list]
    foreach job $jobs {
	if {[string match $commandPattern [lindex $job 4]] == 0} {
	    continue
	}
	lappend rc [lindex $job 0] [lindex $job 2] [lindex $job 3] [lindex $job 4]
    }
    return $rc
}

proc tclcron::schedules {identifier {limit 100} {seconds ""}} {
    variable jobs
    if {$seconds == ""} {
	set seconds [clock seconds]
    }
    set rc [list]
    set id -1
    set i 0; foreach job $jobs {
	if {[lindex $job 0] == $identifier} {
	    set id $i
	    break
	}
	incr i
    }
    
    # no such schedule - TODO: throw an error
    if {$id < 0} {return [list]}
    
    for {set i 0} {$i < $limit} {incr i} {
	if {![tclcron::_schedulenext $id $seconds]} {
	    break
	}
	set next [lindex $jobs $id 1]
	lappend rc [expr {$next+1}]
	set seconds [expr {$next+1}]
    }
    return $rc
}

# protected/private API

#
# re-schedule another execution of a command, and re-initialize waiting event.
#
proc tclcron::_reschedule {id} {
    variable jobs
    _schedulenext $id [clock seconds]
    set jobs [lsort -index 1 -integer $jobs]
    _mkafter
}

#
# cancel currently waiting event,
# if there is any procedure pending to be called, call it.
# (using _call, so that a new waiting event is re-initialized)
#
proc tclcron::_mkafter {} {
    variable jobs
    variable after
    variable maxWait

    after cancel $after
    if {[llength $jobs]>0} {
	set cnext [lindex $jobs 0 1]
	set diff [expr {$cnext-[clock seconds]}]
	if {$diff<=0} {
	    _call 0
	    return
	}
	if {$diff>$maxWait} {set diff $maxWait}
	set after [after [expr {$diff*1000}] tclcron::_mkafter]
    }
}

#
# call a pending function,
# re-schedule waiting event using _reschedule proc.
#
proc tclcron::_call {id} {
    variable jobs
    set job [lindex $jobs $id]
    # try calling the procedure
    # if it fails, throw the error back
    if {[catch {
	uplevel #0 [lindex $job 4]
    } rc]} {
	# throw the error
	if {[catch {
	    bgerror $rc
	}]} {
	    puts stderr "Tclcron background error: $rc"
	}
    }
    # schedule another call of this command
    _reschedule $id
}

#
# calculate next scheduled time a proc is to be executed,
# or delete it from the events queue if it is no longer valid.
#
proc tclcron::_schedulenext {id seconds} {
    variable jobs
    set jar [lindex $jobs $id 3]

    foreach increment {
	1 2 5 10 60 300 600 1200 1800 2400 3600
	7200 14400 21600 43200 86400 129600 172800
	259200 345600 432000 518400 604800 691200
	777600 864000 950400 1036800 1123200 1209600
	1814400 2419200 3024000 3628800 5184000
	7776000 10368000 12960000 15552000 18144000
	20736000 23328000 25920000 28512000 31104000
	31536000 31622400 31968000
    } {
	# try with various date increments

	set current [expr {$seconds + $increment}]
	set next [[lindex $jobs $id 2]::schedulenext jar $current]

	if {$next==""} {
	    set jobs [lreplace $jobs $id $id]
	    return false
	}

	if {$next <= $seconds} {
	    continue
	}
	incr next -1
	lset jobs $id 3 $jar
	lset jobs $id 1 $next
	return true
    }
    if {!$ok} {
	set jobs [lreplace $jobs $id $id]
    }
    return false
}

